# -*- coding: utf-8 -*-
# This file is part of finchan.

# Copyright (C) 2017-present qytz <hhhhhf@foxmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__title__ = 'finchan'
__description__ = 'An event process framework with python'
__url__ = 'https://github.com/qytz/finchan'
__version__ = '0.2.0'
__build__ = 0x000200
__author__ = 'qytz'
__author_email__ = 'hhhhhf@foxmail.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2017-present qytz'
